@@ -57,6 +57,7 @@ pattern RPL_YOURID                  = ReplyCode 042
 pattern RPL_SAVENICK                = ReplyCode 043
 pattern RPL_ATTEMPTINGJUNC          = ReplyCode 050
 pattern RPL_ATTEMPTINGREROUTE       = ReplyCode 051
+pattern RPL_REMOTESUPPORT           = ReplyCode 105 -- ircd-seven unnamed code
 pattern RPL_TRACELINK               = ReplyCode 200
 pattern RPL_TRACECONNECTING         = ReplyCode 201
 pattern RPL_TRACEHANDSHAKE          = ReplyCode 202
@@ -206,7 +207,7 @@ pattern RPL_ENDOFINFO               = ReplyCode 374
 pattern RPL_MOTDSTART               = ReplyCode 375
 pattern RPL_ENDOFMOTD               = ReplyCode 376
 pattern RPL_WHOISHOST               = ReplyCode 378
-pattern RPL_KICKLINKED              = ReplyCode 379
+pattern RPL_WHOISMODES              = ReplyCode 379
 pattern RPL_YOUREOPER               = ReplyCode 381
 pattern RPL_REHASHING               = ReplyCode 382
 pattern RPL_YOURESERVICE            = ReplyCode 383
@@ -402,6 +403,7 @@ pattern ERR_NOPRIVS                 = ReplyCode 723
 pattern RPL_TESTMASK                = ReplyCode 724
 pattern RPL_TESTLINE                = ReplyCode 725
 pattern RPL_NOTESTLINE              = ReplyCode 726
+pattern RPL_TESTMASKGECOS           = ReplyCode 727
 pattern RPL_QUIETLIST               = ReplyCode 728
 pattern RPL_ENDOFQUIETLIST          = ReplyCode 729
 pattern RPL_MONONLINE               = ReplyCode 730
@@ -484,6 +486,7 @@ replyCodeInfoTable
   , (RPL_SAVENICK              , ReplyCodeInfo ClientServerReply "save-nick")
   , (RPL_ATTEMPTINGJUNC        , ReplyCodeInfo ClientServerReply "attempting-junc")
   , (RPL_ATTEMPTINGREROUTE     , ReplyCodeInfo ClientServerReply "attempting-reroute")
+  , (RPL_REMOTESUPPORT         , ReplyCodeInfo ClientServerReply "remote-support")
   , (RPL_TRACELINK             , ReplyCodeInfo CommandReply "trace-link")
   , (RPL_TRACECONNECTING       , ReplyCodeInfo CommandReply "trace-connecting")
   , (RPL_TRACEHANDSHAKE        , ReplyCodeInfo CommandReply "trace-handshake")
@@ -633,7 +636,7 @@ replyCodeInfoTable
   , (RPL_MOTDSTART             , ReplyCodeInfo CommandReply "motd-start")
   , (RPL_ENDOFMOTD             , ReplyCodeInfo CommandReply "end-of-motd")
   , (RPL_WHOISHOST             , ReplyCodeInfo CommandReply "whois-host")
-  , (RPL_KICKLINKED            , ReplyCodeInfo CommandReply "kick-linked")
+  , (RPL_WHOISMODES            , ReplyCodeInfo CommandReply "whois-modes")
   , (RPL_YOUREOPER             , ReplyCodeInfo CommandReply "youre-oper")
   , (RPL_REHASHING             , ReplyCodeInfo CommandReply "rehashing")
   , (RPL_YOURESERVICE          , ReplyCodeInfo CommandReply "youre-service")
@@ -829,6 +832,7 @@ replyCodeInfoTable
   , (RPL_TESTMASK              , ReplyCodeInfo CommandReply "test-mask")
   , (RPL_TESTLINE              , ReplyCodeInfo CommandReply "test-line")
   , (RPL_NOTESTLINE            , ReplyCodeInfo CommandReply "no-test-line")
+  , (RPL_TESTMASKGECOS         , ReplyCodeInfo CommandReply "testmask-gecos")
   , (RPL_QUIETLIST             , ReplyCodeInfo CommandReply "quiet-list")
   , (RPL_ENDOFQUIETLIST        , ReplyCodeInfo CommandReply "end-of-quiet-list")
   , (RPL_MONONLINE             , ReplyCodeInfo CommandReply "mon-online")
