@@ -14,19 +14,23 @@ module Irc.Commands
   , ircCapEnd
   , ircCapLs
   , ircCapReq
+  , ircChantrace
   , ircCnotice
   , ircCprivmsg
+  , ircEtrace
   , ircInfo
   , ircInvite
   , ircIson
   , ircJoin
   , ircKick
   , ircKill
+  , ircKline
   , ircKnock
   , ircLinks
   , ircList
   , ircLusers
   , ircMap
+  , ircMasktrace
   , ircMode
   , ircMotd
   , ircNick
@@ -37,33 +41,40 @@ module Irc.Commands
   , ircPing
   , ircPong
   , ircPrivmsg
-  , ircRules
   , ircQuit
   , ircRemove
+  , ircRules
   , ircStats
+  , ircTestline
+  , ircTestmask
   , ircTime
   , ircTopic
+  , ircUnkline
   , ircUser
   , ircUserhost
+  , ircVersion
   , ircWho
   , ircWhois
   , ircWhowas
-  , ircVersion
 
   -- * ZNC support
   , ircZnc
 
   -- * SASL support
+  , AuthenticatePayload(..)
   , ircAuthenticate
-  , plainAuthenticationMode
+  , ircAuthenticates
   , encodePlainAuthentication
+  , encodeExternalAuthentication
   ) where
 
 import           Irc.RawIrcMsg
 import           Irc.Identifier
+import           Data.ByteString (ByteString)
 import           Data.Text (Text)
 import qualified Data.Text as Text
 import qualified Data.Text.Encoding as Text
+import qualified Data.ByteString as B
 import qualified Data.ByteString.Base64 as Enc
 
 nonempty :: Text -> [Text]
@@ -193,6 +204,52 @@ ircKill ::
   RawIrcMsg
 ircKill who msg = rawIrcMsg "KILL" (who : nonempty msg)
 
+-- | KLINE command
+ircKline ::
+  Text {- ^ minutes -} ->
+  Text {- ^ mask    -} ->
+  Text {- ^ reason  -} ->
+  RawIrcMsg
+ircKline minutes mask reason = rawIrcMsg "KLINE" [minutes, mask, reason]
+
+-- | UNKLINE command
+ircUnkline ::
+  Text {- ^ mask -} ->
+  RawIrcMsg
+ircUnkline mask = rawIrcMsg "UNKLINE" [mask]
+
+-- | TESTLINE command
+ircTestline ::
+  Text {- ^ mask -} ->
+  RawIrcMsg
+ircTestline mask = rawIrcMsg "TESTLINE" [mask]
+
+-- | TESTMASK command
+ircTestmask ::
+  Text {- ^ mask  -} ->
+  Text {- ^ gecos -} ->
+  RawIrcMsg
+ircTestmask mask gecos = rawIrcMsg "TESTMASK" (mask : nonempty gecos)
+
+-- | MASKTRACE command
+ircMasktrace ::
+  Text {- ^ mask  -} ->
+  Text {- ^ gecos -} ->
+  RawIrcMsg
+ircMasktrace mask gecos = rawIrcMsg "MASKTRACE" [mask, gecos]
+
+-- | CHANTRACE command
+ircChantrace ::
+  Text {- ^ channel -} ->
+  RawIrcMsg
+ircChantrace channel = rawIrcMsg "CHANTRACE" [channel]
+
+-- | ETRACE command
+ircEtrace ::
+  Text {- ^ argument -} ->
+  RawIrcMsg
+ircEtrace arg = rawIrcMsg "ETRACE" [arg]
+
 -- | REMOVE command
 ircRemove ::
   Identifier {- ^ channel  -} ->
@@ -329,15 +386,8 @@ ircAdmin = rawIrcMsg "ADMIN" . nonempty
 -- | USER command
 ircUser ::
   Text {- ^ username -} ->
-  Bool {- ^ set +w   -} ->
-  Bool {- ^ set +i   -} ->
   Text {- ^ realname -} -> RawIrcMsg
-ircUser user set_w set_i real = rawIrcMsg "USER" [user, modeTxt, "*", real]
-  where
-    modeTxt = Text.pack (show mode)
-    mode :: Int
-    mode = (if set_w then 4 else 0) -- bit 2
-         + (if set_i then 8 else 0) -- bit 3
+ircUser user real = rawIrcMsg "USER" [user, "0", "*", real]
 
 -- | CAP REQ command
 ircCapReq ::
@@ -349,9 +399,9 @@ ircCapReq caps = rawIrcMsg "CAP" ["REQ", Text.unwords caps]
 ircCapEnd :: RawIrcMsg
 ircCapEnd = rawIrcMsg "CAP" ["END"]
 
--- | CAP LS command
+-- | CAP LS command - support CAP version 3.2
 ircCapLs :: RawIrcMsg
-ircCapLs = rawIrcMsg "CAP" ["LS"]
+ircCapLs = rawIrcMsg "CAP" ["LS", "302"]
 
 -- | ZNC command
 --
@@ -361,21 +411,43 @@ ircZnc ::
   RawIrcMsg
 ircZnc = rawIrcMsg "ZNC"
 
--- | AUTHENTICATE command
-ircAuthenticate :: Text -> RawIrcMsg
+-- | Payload for 'ircAuthenticates'
+newtype AuthenticatePayload = AuthenticatePayload ByteString
+  deriving Show
+
+ircAuthenticate ::
+  Text {- ^ authentication mechanism -} ->
+  RawIrcMsg
 ircAuthenticate msg = rawIrcMsg "AUTHENTICATE" [msg]
 
--- | PLAIN authentiation mode
-plainAuthenticationMode :: Text
-plainAuthenticationMode = "PLAIN"
+-- | AUTHENTICATE command generator. Returns a list
+-- because AUTHENTICATE has a chunking behavior.
+ircAuthenticates ::
+  AuthenticatePayload {- ^ authentication payload -} ->
+  [RawIrcMsg]
+ircAuthenticates (AuthenticatePayload bytes) =
+  map (ircAuthenticate . Text.decodeUtf8) (chunks (Enc.encode bytes))
+  where
+    chunks :: ByteString -> [ByteString]
+    chunks b
+      | B.null b          = ["+"]
+      | B.length b >= 400 = B.take 400 b : chunks (B.drop 400 b)
+      | otherwise         = [b]
 
 -- | Encoding of username and password in PLAIN authentication
 encodePlainAuthentication ::
   Text {- ^ username -} ->
   Text {- ^ password -} ->
-  Text
+  AuthenticatePayload
 encodePlainAuthentication user pass
-  = Text.decodeUtf8
-  $ Enc.encode
+  = AuthenticatePayload
   $ Text.encodeUtf8
   $ Text.intercalate "\0" [user,user,pass]
+
+-- | Encoding of username in EXTERNAL authentication
+encodeExternalAuthentication ::
+  Text {- ^ username -} ->
+  AuthenticatePayload
+encodeExternalAuthentication
+  = AuthenticatePayload
+  . Text.encodeUtf8
