@@ -17,6 +17,7 @@ module Irc.Message
   -- * High-level messages
     IrcMsg(..)
   , CapCmd(..)
+  , CapMore(..)
   , cookIrcMsg
 
   -- * Properties of messages
@@ -28,6 +29,7 @@ module Irc.Message
   -- * Helper functions
   , nickSplit
   , computeMaxMessageLength
+  , capCmdText
   ) where
 
 import           Control.Monad
@@ -47,7 +49,7 @@ data IrcMsg
   = UnknownMsg !RawIrcMsg -- ^ pass-through for unhandled messages
   | Reply !ReplyCode [Text] -- ^ code arguments
   | Nick !UserInfo !Identifier -- ^ old new
-  | Join !UserInfo !Identifier -- ^ user channel
+  | Join !UserInfo !Identifier !Text -- ^ user channel account(extended-join)
   | Part !UserInfo !Identifier (Maybe Text) -- ^ user channel reason
   | Quit !UserInfo (Maybe Text) -- ^ user reason
   | Kick !UserInfo !Identifier !Identifier !Text -- ^ kicker channel kickee comment
@@ -58,33 +60,42 @@ data IrcMsg
   | Notice !UserInfo !Identifier !Text -- ^ source target txt
   | Mode !UserInfo !Identifier [Text] -- ^ source target txt
   | Authenticate !Text -- ^ parameters
-  | Cap !CapCmd [Text] -- ^ command parameters
+  | Cap !CapCmd -- ^ cap command and parameters
   | Ping [Text] -- ^ parameters
   | Pong [Text] -- ^ parameters
   | Error !Text -- ^ message
-  | BatchStart Text Text [Text] -- ^ reference-id type parameters
-  | BatchEnd   Text             -- ^ reference-id
+  | BatchStart !Text !Text [Text] -- ^ reference-id type parameters
+  | BatchEnd !Text -- ^ reference-id
+  | Account !UserInfo !Text -- ^ user account name changed (account-notify extension)
+  | Chghost !UserInfo !Text !Text -- ^ Target, new username and new hostname
+  | Wallops  !UserInfo !Text -- ^ Braodcast message: Source, message
   deriving Show
 
--- | Sub-commands of the CAP command
+data CapMore = CapMore | CapDone
+  deriving (Show, Read, Eq, Ord)
+
+-- | Sub-commands of the CAP command sent by server
 data CapCmd
-  = CapLs -- ^ request list of supported caps
-  | CapList -- ^ request list of active caps
-  | CapReq -- ^ request activation of cap
-  | CapAck -- ^ request accepted
-  | CapNak -- ^ request denied
-  | CapEnd -- ^ end negotiation
-  deriving (Show, Eq, Ord)
+  = CapLs !CapMore [(Text, Maybe Text)] -- ^ list of supported caps
+  | CapList [Text] -- ^ list of active caps
+  | CapAck [Text] -- ^ request accepted
+  | CapNak [Text] -- ^ request denied
+  | CapNew [(Text, Maybe Text)] -- ^ new capability available (cap-notify extension)
+  | CapDel [Text] -- ^ capability removed (cap-notify extension)
+  deriving (Show, Read, Eq, Ord)
 
 -- | Match command text to structured cap sub-command
-cookCapCmd :: Text -> Maybe CapCmd
-cookCapCmd "LS"   = Just CapLs
-cookCapCmd "LIST" = Just CapList
-cookCapCmd "ACK"  = Just CapAck
-cookCapCmd "NAK"  = Just CapNak
-cookCapCmd "END"  = Just CapEnd
-cookCapCmd "REQ"  = Just CapReq
-cookCapCmd _      = Nothing
+cookCapCmd :: Text -> [Text] -> Maybe CapCmd
+cookCapCmd cmd args =
+  case (cmd, args) of
+    ("LS"  , ["*", caps]) -> Just (CapLs CapMore (splitCapList caps))
+    ("LS"  , [     caps]) -> Just (CapLs CapDone (splitCapList caps))
+    ("LIST", [     caps]) -> Just (CapList (Text.words caps))
+    ("ACK" , [     caps]) -> Just (CapAck (Text.words caps))
+    ("NAK" , [     caps]) -> Just (CapNak (Text.words caps))
+    ("NEW" , [     caps]) -> Just (CapNew (splitCapList caps))
+    ("DEL" , [     caps]) -> Just (CapDel (Text.words caps))
+    _                     -> Nothing
 
 -- | Interpret a low-level 'RawIrcMsg' as a high-level 'IrcMsg'.
 -- Messages that can't be understood are wrapped in 'UnknownMsg'.
@@ -94,8 +105,7 @@ cookIrcMsg msg =
     cmd | Right (n,"") <- decimal cmd ->
         Reply (ReplyCode n) (view msgParams msg)
     "CAP" | _target:cmdTxt:rest <- view msgParams msg
-          , Just cmd <- cookCapCmd cmdTxt ->
-           Cap cmd rest
+          , Just cmd <- cookCapCmd cmdTxt rest -> Cap cmd
 
     "AUTHENTICATE" | x:_ <- view msgParams msg ->
         Authenticate x
@@ -118,9 +128,13 @@ cookIrcMsg msg =
              Nothing         -> Notice user (mkId chan) txt
 
     "JOIN" | Just user <- view msgPrefix msg
-           , chan:_    <- view msgParams msg ->
+           , chan:rest <- view msgParams msg ->
 
-           Join user (mkId chan)
+              Join user (mkId chan)
+            $ case rest of
+                ["*" , _real] -> ""
+                [acct, _real] -> acct
+                _             -> ""
 
     "QUIT" | Just user <- view msgPrefix msg
            , reasons   <- view msgParams msg ->
@@ -157,6 +171,18 @@ cookIrcMsg msg =
             , Just ('-',refid') <- Text.uncons refid ->
             BatchEnd refid'
 
+    "ACCOUNT" | Just user <- view msgPrefix msg
+              , [acct] <- view msgParams msg ->
+      Account user (if acct == "*" then "" else acct)
+
+    "CHGHOST" | Just user <- view msgPrefix msg
+              , [newuser, newhost] <- view msgParams msg ->
+      Chghost user newuser newhost
+
+    "WALLOPS" | Just user <- view msgPrefix msg
+              , [txt] <- view msgParams msg ->
+      Wallops user txt
+
     _      -> UnknownMsg msg
 
 -- | Parse a CTCP encoded message:
@@ -176,37 +202,46 @@ data MessageTarget
   | TargetWindow !Identifier -- ^ Directed message to channel or from user
   | TargetNetwork            -- ^ Network-level message
   | TargetHidden             -- ^ Completely hidden message
+  deriving (Show)
 
 -- | Target information for the window that could be appropriate to
 -- display this message in.
 msgTarget :: Identifier -> IrcMsg -> MessageTarget
 msgTarget me msg =
   case msg of
-    UnknownMsg{}                  -> TargetNetwork
-    Nick user _                   -> TargetUser (userNick user)
-    Mode _ tgt _ | tgt == me      -> TargetNetwork
-                 | otherwise      -> TargetWindow tgt
-    Join _ chan                   -> TargetWindow chan
-    Part _ chan _                 -> TargetWindow chan
-    Quit user _                   -> TargetUser (userNick user)
-    Kick _ chan _ _               -> TargetWindow chan
-    Topic _ chan _                -> TargetWindow chan
-    Privmsg src tgt _ | tgt == me -> TargetWindow (userNick src)
-                      | otherwise -> TargetWindow tgt
-    Ctcp src tgt _ _  | tgt == me -> TargetWindow (userNick src)
-                      | otherwise -> TargetWindow tgt
-    CtcpNotice src tgt _ _  | tgt == me -> TargetWindow (userNick src)
-                            | otherwise -> TargetWindow tgt
-    Notice  src tgt _ | tgt == me -> TargetWindow (userNick src)
-                      | otherwise -> TargetWindow tgt
-    Authenticate{}                -> TargetHidden
-    Ping{}                        -> TargetHidden
-    Pong{}                        -> TargetHidden
-    Error{}                       -> TargetNetwork
-    Cap{}                         -> TargetNetwork
-    Reply{}                       -> TargetNetwork
-    BatchStart{}                  -> TargetHidden
-    BatchEnd{}                    -> TargetHidden
+    UnknownMsg{}             -> TargetNetwork
+    Nick user _              -> TargetUser (userNick user)
+    Mode _ tgt _ | tgt == me -> TargetNetwork
+                 | otherwise -> TargetWindow tgt
+    Join _ chan _            -> TargetWindow chan
+    Part _ chan _            -> TargetWindow chan
+    Quit user _              -> TargetUser (userNick user)
+    Kick _ chan _ _          -> TargetWindow chan
+    Topic _ chan _           -> TargetWindow chan
+    Privmsg src tgt _        -> directed src tgt
+    Ctcp src tgt _ _         -> directed src tgt
+    CtcpNotice src tgt _ _   -> directed src tgt
+    Notice  src tgt _        -> directed src tgt
+    Authenticate{}           -> TargetHidden
+    Ping{}                   -> TargetHidden
+    Pong{}                   -> TargetHidden
+    Error{}                  -> TargetNetwork
+    Cap{}                    -> TargetNetwork
+    Reply code args          -> replyTarget code args
+    BatchStart{}             -> TargetHidden
+    BatchEnd{}               -> TargetHidden
+    Account user _           -> TargetUser (userNick user)
+    Chghost user _ _         -> TargetUser (userNick user)
+    Wallops src _            -> TargetWindow (userNick src)
+  where
+    directed src tgt
+      | Text.null (userHost src) = TargetNetwork -- server message
+      | tgt == me = TargetWindow (userNick src)
+      | otherwise = TargetWindow tgt
+
+    replyTarget RPL_TOPIC    (_:chan:_)   = TargetWindow (mkId chan)
+    replyTarget RPL_INVITING (_:_:chan:_) = TargetWindow (mkId chan)
+    replyTarget _                _        = TargetNetwork
 
 -- | 'UserInfo' of the user responsible for a message.
 msgActor :: IrcMsg -> Maybe UserInfo
@@ -215,7 +250,7 @@ msgActor msg =
     UnknownMsg{}  -> Nothing
     Reply{}       -> Nothing
     Nick x _      -> Just x
-    Join x _      -> Just x
+    Join x _ _    -> Just x
     Part x _ _    -> Just x
     Quit x _      -> Just x
     Kick x _ _ _  -> Just x
@@ -225,6 +260,7 @@ msgActor msg =
     CtcpNotice x _ _ _ -> Just x
     Notice x _ _  -> Just x
     Mode x _ _    -> Just x
+    Account x _   -> Just x
     Authenticate{}-> Nothing
     Ping{}        -> Nothing
     Pong{}        -> Nothing
@@ -232,6 +268,8 @@ msgActor msg =
     Cap{}         -> Nothing
     BatchStart{}  -> Nothing
     BatchEnd{}    -> Nothing
+    Chghost x _ _ -> Just x
+    Wallops x _   -> Just x
 
 -- | Text representation of an IRC message to be used for matching with
 -- regular expressions.
@@ -241,7 +279,7 @@ ircMsgText msg =
     UnknownMsg raw -> Text.unwords (view msgCommand raw : view msgParams raw)
     Reply (ReplyCode n) xs -> Text.unwords (Text.pack (show n) : xs)
     Nick x y       -> Text.unwords [renderUserInfo x, idText y]
-    Join x _       -> renderUserInfo x
+    Join x _ _     -> renderUserInfo x
     Part x _ mb    -> Text.unwords (renderUserInfo x : maybeToList mb)
     Quit x mb      -> Text.unwords (renderUserInfo x : maybeToList mb)
     Kick x _ z r   -> Text.unwords [renderUserInfo x, idText z, r]
@@ -253,11 +291,31 @@ ircMsgText msg =
     Mode x _ xs    -> Text.unwords (renderUserInfo x:"set mode":xs)
     Ping xs        -> Text.unwords xs
     Pong xs        -> Text.unwords xs
-    Cap _ xs       -> Text.unwords xs
+    Cap cmd        -> capCmdText cmd
     Error t        -> t
+    Account x a    -> Text.unwords [renderUserInfo x, a]
     Authenticate{} -> ""
     BatchStart{}   -> ""
     BatchEnd{}     -> ""
+    Chghost x a b  -> Text.unwords [renderUserInfo x, a, b]
+    Wallops x t    -> Text.unwords [renderUserInfo x, t]
+
+capCmdText :: CapCmd -> Text
+capCmdText cmd =
+  case cmd of
+    CapLs more caps -> capMoreText more <> capUnsplitCaps caps
+    CapNew     caps -> capUnsplitCaps caps
+    CapList    caps -> Text.unwords caps
+    CapAck     caps -> Text.unwords caps
+    CapNak     caps -> Text.unwords caps
+    CapDel     caps -> Text.unwords caps
+
+capMoreText :: CapMore -> Text
+capMoreText CapDone = ""
+capMoreText CapMore = "* "
+
+capUnsplitCaps :: [(Text, Maybe Text)] -> Text
+capUnsplitCaps xs = Text.unwords [ k <> maybe "" ("=" <>) v | (k, v) <- xs ]
 
 -- nickname   =  ( letter / special ) *8( letter / digit / special / "-" )
 -- letter     =  %x41-5A / %x61-7A       ; A-Z / a-z
@@ -291,3 +349,12 @@ computeMaxMessageLength myUserInfo target
   - Text.length (renderUserInfo myUserInfo)
   - length (": PRIVMSG  :\r\n"::String)
   - Text.length target
+
+splitCapList :: Text -> [(Text, Maybe Text)]
+splitCapList caps =
+  [ (name, value)
+    | kv <- Text.words caps
+    , let (name, v) = Text.break ('=' ==) kv
+    , let value | Text.null v = Nothing
+                | otherwise   = Just $! Text.tail v
+    ]
